//====================================================================//
//                 tetgame.c       20/12/93      Michael Glickman     //
//====================================================================//
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include "tetX.h"
#include "tet_values.h"
#include "keyjoy.h"
#include "kbcodes.h"

#define drawfullboard DrawBoard(Board, 0, BRD_WIDTH, 0, BRD_HEIGHT)

/* ******************************************************************
//                        GLOBAL VARIABLES
 * ******************************************************************/
BOARD  Board[BRD_HEIGHT][BRD_WIDTH];

FIGURE  NexFig;

int  DefLevel, DefLays;         // Default level and number of layers
int  JoyEquiv[4];               // Equivalent krys for joystick buttons
int  Level, Layers;             // Current level and number of layers
int  NexTyp, NexRot;            // Next type and rotation
int  FullLays;                  // Number of layers to be filled prior to game
int  Moves;                     // Number of figures presented
unsigned int  Score;            // Current score
unsigned int  Bonus;            // Current bonus value;

int  NewLevelAt;                // Score where new level arrives
int  JoyEnabled;                // True if joystick is enabled
int  Demo;                      // True for demo game
int  ShowGrid;                  // True, if grid has to be shown
int  ShowNext;                  // True, if next figure has to be shown
int  DelayMove, DelayJoy;       // Delay before moving figure or
                                // checking joystick

long AdvTicks;                  // Ticks when figure should be advanced
long AdvJoy;                    // Ticks when joystick should be checked


//  Statistics
unsigned int  Scores[10];        // Scores for each level
int     Statist[NFIGS];          // Statistics by number of figures
int     StatLays[10];            // Layers for each level


/* ******************************************************************
//                       EXTERNAL VARIABLES
// ******************************************************************/
extern long  Ticks;                  // Timer ticks from program start
extern int   BoardX, BoardY;         // Board position
extern int   BoardW, BoardH;         // Board size in pixels
extern int   DefLevel, DefLays;      // Starting level and layers
extern int   NextX, NextY;           // Position of 'Next' window
extern char  Arrows[4][8][2];
extern char  *PlayListFileName;
extern int   FigTurns[];
extern int   MusicEnabled;
extern int   ScrollAvailable;
extern BOTTOM_MODE BottomMode;

extern int   BeepDown[], BeepDrop[], BeepLayer[], BeepLevel[];
extern int   BeepBottom[], BeepSkip[];

unsigned int PrevScore; // Score before start of current level
int  PrevLays;          // Full layers before start of current level
/* ****************************************************************
                        LOCAL VARIABLES
****************************************************************** */
static int  OldX, OldY;         // Position of figure before move


/* ****************************************************************** */
/*                      EXECUTION CODE                                */
/* *****************************************************************  */

/* ---------------------------------------------------------------- */
/*                       Figure moving                              */
/* ---------------------------------------------------------------- */
static void shift_figure(FIGURE figure, int x, int y, int wid, int hei)
        {   BOARD  EmbBoard[BRD_HEIGHT][BRD_WIDTH];

    int x1, x2, y1, y2, j;


    if (x < OldX) { x1 = x; x2 = OldX; }
          else    { x1 = OldX; x2 = x; }

    if (x1 <0) x1 = 0;
    x2 += FIG_DIM;
    if (x2 > BRD_WIDTH) x2 = BRD_WIDTH;

    y1 = OldY;
    y2 = y + FIG_DIM;
    if (y2 > BRD_HEIGHT) y2 = BRD_HEIGHT;

    for (j=y1; j<y2; j++)
        memcpy( EmbBoard[j],  Board[j], BRD_WIDTH*sizeof(BOARD));

    embed_figure(EmbBoard, figure, x, y, wid, hei);

    DrawBoard(EmbBoard, x1, x2, y1, y2);
#ifdef X11
    if (BottomMode == B_SHADOW)
#else	
    if (!Demo && BottomMode == B_SHADOW)
#endif
              ShowShadow(figure, x);

    adjusteye(x, y);
}


static int drop_figure(FIGURE figure, int x, int y, int wid, int hei)
{ int y1;

     for (y1=y; test_posit(figure, x, y1, wid, hei); y1++);


     return y1-1;

}

static int squeeze(long curtim)
{  register int x, y, j, sq;
   int  y1;
   int  sqarr[FIG_DIM];

   BOARD Brd1[BRD_HEIGHT][BRD_WIDTH];


   /* Find first nonempty row */

   for (y=0; y<BRD_HEIGHT; y++)
   {  for (x = 0;  x<BRD_WIDTH && Board[y][x] == 0; x++);
       if (x<BRD_WIDTH) break;
   }

   y1 = y;

/* Examine for full rows */
   for (sq=j=0; y<BRD_HEIGHT; y++)
   {

     for (x = 0;  x<BRD_WIDTH && Board[y][x]; x++); /* Verify on full row */

     if (x>=BRD_WIDTH) sqarr[sq++] = y;            /* mark full row */
     else memcpy(Brd1[j++], Board[y], BRD_WIDTH*sizeof(BOARD));


   }


   if (sq > 0)
   {   RECT rect;
      
       if (!Demo) owlwave(2);

#define REPEAT1 2
#define REPEAT2 3

//     if (SoundType & SOUND_BEEP)
//         music_start(BeepLayer, REPEAT1);
      
      rect.left = BoardX;
      rect.right = BoardX + BoardW;
      rect.top = BoardY + FIG_HEIGHT * sqarr[0];
      rect.bottom = BoardY + FIG_HEIGHT * (sqarr[sq-1] + 1);

      for (x=0; x<REPEAT2; x++)
      {  if (x>0) RefreshAndWait(1, &rect); // mus_delay(1);

        
         for (y=0; y<sq; y++)
           { j = sqarr[y];
             DrawBoardRow(Board[j], 0, BRD_WIDTH, j, 0xff);
           }

         RefreshAndWait(3, &rect);

         for (y=0; y<sq; y++)
           { j = sqarr[y];
             DrawBoardRow(Board[j], 0, BRD_WIDTH, j, 0x0);
           }

      }

      j = y1+sq;
      for (y=BRD_HEIGHT-1; y>=j; y--)
         memcpy( Board[y], Brd1[y-j], BRD_WIDTH*sizeof(BOARD));

      while (y >= y1)
         memset( Board[y--], 0, BRD_WIDTH*sizeof(BOARD));


      Layers+=sq;

      DrawBoard(Board, 0, BRD_WIDTH, y1, sqarr[sq-1]+1);


/*      if (SoundType == SOUND_BEEP)
           while(music_ctrl(0))
             if (Demo||ShareWare) movlet_ctrl(0);
*/
   }  /* if (sq>0) */



   return sq;
}




/*
static void Shell(void)
{
  int cpid;
  char *ShellName = getenv("SHELL");
  char *ShortShellName = strrchr(ShellName, '/');

  if (!ShortShellName) ShortShellName = ShellName;
           else        ShortShellName++;

  SuspendTimer();
  vga_flip();

  cpid = fork();
  if (cpid==0)
  {     execle(ShellName, ShortShellName, NULL);
        exit(0);
  }

  waitpid(cpid, NULL, 0);
  kill(cpid, SIGTERM);

  vga_flip();
  ResumeTimer();

  return;

}
*/



// Get command, by keyboard or joystick
int getCommand(void)
{  long CurTime = Ticks;
   int key = 0;

   if ( JoyEnabled && CurTime > AdvJoy)
   {   key = getKeyJoy();

       if (key > JOY_START && key <= JOY_LAST)
       {
       
           ShowJoyPos(key - JOY_ARROWS);
           AdvJoy = CurTime + DelayJoy;

	   if (key < JOY_START + 16)
	   {  int i, mask;
	       for (i=0, mask = 1; i<4; i++, mask <<= 1)         
	       {  if (key & mask)
	          {  key = JoyEquiv[i];
		     break;
		  }
	       }	       
	   }   
	   
       }
       else
          ShowJoyPos(0);

   }
   else
       key = kbdGet();

   return key;
}


static int down_figure(FIGURE figure, int x, int y, int wid, int hei)
{
    if ( test_posit (figure, x, ++y, wid, hei))
    {  if (y%2 && Bonus>0) {Bonus --; ShowBonus();}
//       if ((SoundType & SOUND_BEEP) && !music_play_stat())
/*       if ((SoundType & SOUND_BEEP))
                          music_start(BeepDown, 1);  */
       return 1;
    }

    return 0;
}


#define LOSE_ON_SKIP 50
int move_figure(void)
{    int x, y, t, ok, shif;
     int wid, hei;
     long    curtim;
     FIGURE    figure, figure1;
 // static int Keys[5] = {0, '4', '5', '6', '2'};

     cop_figure(figure, NexFig);
     Statist[NexTyp] += 1;

     wid = fig_width(figure);
     hei = fig_height(figure);
     x = (BRD_WIDTH-wid) >> 1; y = 0;
     ok = test_posit(figure, x, y, wid, hei);
     if (!ok) return 8;

 //    owlwave(0);     /* Wings down, mouth close; */

     Bonus = BRD_HEIGHT;
     if (ShowNext) Bonus -= 5;
     if (ShowGrid) Bonus -= 2;
     if (BottomMode == B_SHADOW) Bonus -= 3;
     if (Bonus < 15) Bonus = 15;

     ShowBonus();

     NexTyp = get_rand(NFIGS);                   /* Next figure: type */
     NexRot = get_rand(FigTurns[NexTyp]);        /*          rotation */

     get_figure(NexFig, NexTyp, NexRot);

     if (ShowNext) ShowNextFigure();

     OldY = y;  OldX = x;
     shift_figure(figure, x, y, wid, hei);


     curtim = Ticks;
     AdvTicks = curtim + DelayMove;
     AdvJoy = curtim + DelayJoy;

/*     while (kbhit()) getch(); */


     while (ok)
      {  OldX = x;  OldY = y;

        t = getCommand();
        if (t < 0x100) t = toupper(t);

        shif = 0;

        switch(t)
        {  case '4':          /* Shift left */
           case KY_LEFT:
           case JOY_LEFT:
             t = x - 1;
             shif = test_posit (figure, t, y, wid, hei);
             if (shif) x=t;
             break;

           case '6':          /* Shift right */
           case KY_RIGHT:
           case JOY_RIGHT:
             t = x + 1;
             shif = test_posit (figure, t, y, wid, hei);
             if (shif) x=t;
             break;

           case '8':          /* Rotate right */
           case KY_UP:
           case JOY_UP:
             rot_right(figure, figure1);
             if (x+hei > BRD_WIDTH) x = BRD_WIDTH - hei;
             if ( test_posit (figure1, x, y, hei, wid))
                  { shif = 1; cop_figure(figure, figure1);
                    t = wid; wid = hei; hei = t;
                  }
             break;


           case '5':          /* Rotate left */
           case KY_CENTER:
           case KY_ENTER:
             rot_left(figure, figure1);
             if (x+hei > BRD_WIDTH) x = BRD_WIDTH - hei;
             if ( test_posit (figure1, x, y, hei, wid))
                  { shif = 1; cop_figure(figure, figure1);
                    t = wid; wid = hei; hei = t;
                  }
             break;

                              /* Down */
           case '2':
           case KY_DOWN:
           case JOY_DOWN:
             if(down_figure(figure, x, y, wid, hei))
                  { shif = 1; y++; }
             break;

           case ' ':          /* Drop  */
           case '0':
           case KY_INSERT:
/*           if (SoundType & SOUND_BEEP)
                      music_start(BeepDrop, 1); */
             DrawBoard(Board, x, x+wid, y, y+hei);
             y = drop_figure(figure, x, y, wid, hei);
             adjusteye(x, y);
             owlwave(1);    /* Wings up; */
             ok = 0;
             break;

//
//         case '7':          /* Skip figure */
//         case KY_HOME:
//         if (Score >= LOSE_ON_SKIP)
//           {
//     /*          if (SoundType & SOUND_BEEP)
//                     music_start(BeepSkip, 1);
//               drawfullboard;  */
//               Score -= LOSE_ON_SKIP;
//               ok = 0;
//               y = -1;
//           }
//           break;
//

          case KY_BACK:         /* Terminate */
          case KY_DELETE:
          case 'Q':
             return 8;


           case 'N':          /* Show next */
           case '1':
             ShowNext = ! ShowNext;
             PrepareNextWindow();
             break;

           case '3':          /* Show grid  */
           case 'G':
             ShowGrid = !ShowGrid;
             PrepareShowGrid(figure, x, y, wid, hei);
             break;

           case 'L':          /* Next level */
             if (Level < 9)
                 ChangeLevel(figure, x, y, wid, hei);
             break;


           case 'J':          /* Enable joystick */
			 if (EnableJoystick())
             {  PrepareJoyWindow();
                AdvJoy = curtim;
             }
             break;


           case 'K':          /* Disable joystick */
			 DisableJoystick();
             PrepareJoyWindow();
             break;

           case 'H':          /* Help */
           case KY_F1:
             MaintainHelp();
             break;

           case KY_ESC:        /* Pause */
           case 'P':
              PauseGame();
		      AdvTicks = Ticks + DelayMove;
              break;

/*           case 'D':          // Shell
              Shell();
              break;
*/
           case 'M':            /* Toggle background music */
              if (PlayListFileName)
              { MusicEnabled = !MusicEnabled;
                PrepareMusic(MODE_FORCE);
              }
              break;

/*
           case 'S':          // Toggle sound
               break;
*/

           case 'B':
              ToggleBottomMode();
              break;

         }  /* switch; */


	 curtim = Ticks;

         if (ok)
         {
           if (curtim >= AdvTicks)
           {
               ok = shif = down_figure(figure, x, y, wid, hei);
               if (ok)
                { y++;
                  AdvTicks += DelayMove;
                }
             /*  else
               if (SoundType & SOUND_BEEP)
                  music_start(BeepBottom, 1); */
            }


           if (shif)
               shift_figure(figure, x, y, wid, hei);


         }   /* if (ok) */

         if (BottomMode == B_HELP) ChangeHelpPage();
 

      } /* while (ok)  */



      if (y>=0)
      {   embed_figure(Board, figure, x, y, wid, hei);
                              /* t = last line number */
          DrawBoard(Board, 0, BRD_WIDTH, y, y+hei);

          if (BottomMode == B_SHADOW)
                      ShowShadow(NULL, 0);

          t = squeeze(curtim);
          Score += Bonus + 50*t;

          AdvJoy += DelayJoy;

      }

      if (BottomMode == B_HELP) ChangeHelpPage();
      Moves++;
      
      if (Level < 9 && Moves >= NewLevelAt)
                  ChangeLevel(figure, -1, -1, -1, -1);

      UpdateStatus();

      return 0;
}


int move_demo_figure(void)
{   int x, y, x1, y1, ok, t, turn, shif, key, downc;
    int hei, wid, status, wid1, hei1;
    int BestX, BestT, MaxTurn, BestP;
    int cost, BestCost;
    int delthink, nexthink;
    static int diaps[] = {  4,  3,  3,  3,  2,  2,  2,  1,  1,  1};
    int  fig_y[BRD_WIDTH];

    long   Ticks66 = -1;
    long   ranval;
    int    diap, rv;
    long   curtim;
    FIGURE    figure, figure1, figure2;


     cop_figure(figure, NexFig);
     Statist[NexTyp] += 1;
     MaxTurn = FigTurns[NexTyp];

     wid = fig_width(figure);

     hei = fig_height(figure);

     x = (BRD_WIDTH-wid)>>1; y = 0;
     ok = test_posit( figure, x, y, wid, hei);
     if (!ok) return 4;

     OldX = x; OldY = y;

     NexTyp = get_rand(NFIGS);                 /* Next figure: type     */
     NexRot = get_rand(FigTurns[NexTyp]);      /*              rotation */
     get_figure(NexFig, NexTyp, NexRot);


     curtim = Ticks;
     AdvTicks = curtim + DelayMove;

     delthink = DelayMove>>1;
     if (delthink==0) delthink = 1;
     nexthink = curtim + delthink;


     /* Calculate best position */
     BestX = -1; BestT = 0; BestP = 0;
     BestCost = MAXINT;
     cop_figure(figure1, figure);


     for (turn = 0; turn < MaxTurn; turn ++)
     {

           ShowJoyPos(0);

           if ((turn & 1) == 0)
              { wid1 = wid; hei1 = hei; }
           else
              { wid1 = hei; hei1 = wid; }

	   x1 = BRD_WIDTH-wid1+1;
           while(--x1 >=0)
             fig_y[x1] = drop_figure(figure1, x1, y, wid1, hei1);
	
	   x1 = BRD_WIDTH-wid1+1;
           while (--x1 >= 0)
           { y1 = fig_y[x1];
             cost = calc_cost(figure1, x1, y1, wid1, hei1);
	     t = 0;	
             if (x1 < BRD_WIDTH-wid1 && y1>fig_y[x1+1] &&
		     test_posit(figure1, x1+1, y1, wid1, hei1))
             {    cost -= 100;  t = 1; }
	     else
             if (x1 > 0 && y1>fig_y[x1-1] &&
		     test_posit(figure1, x1-1, y1, wid1, hei1))
             {    cost -= 100;  t = -1; }

             if ((cost < BestCost) || (cost==BestCost && get_rand1(2)))
                   {  BestCost = cost; BestP = t;
                      BestX = x1; BestT = turn;
                   }

           }


           rot_left (figure1, figure2);
           cop_figure(figure1, figure2);

        }



/* Move demo figure */
     //ranval = rand();
     //ranval = (ranval << 16) | rand();
     ranval = random();

     diap = diaps[Level];

     if (ShowNext) ShowNextFigure();

     shift_figure(figure, x, y, wid, hei);

     curtim = Ticks;
     AdvTicks = curtim + DelayMove;
     AdvJoy = curtim + DelayJoy;

     Bonus = BRD_HEIGHT;
     if (ShowNext) Bonus -= 5;
     else
     if (ShowGrid) Bonus -= 2;
  //   if (Bonus < 15) Bonus = 15;
     ShowBonus();

     turn = 0;

     status = 0;

     robthink(status);

     do
     {  OldX = x; OldY = y;
        shif = 0;
        ok = 1;
	downc = 0;

        /* pause(); */

        t = getCommand();
        if (t < 0x100) t = toupper(t);

        if (t=='N') t =0;
        if (t == 0 && Ticks != Ticks66 && Ticks % 666 == 0) t='N';
        Ticks66 = Ticks;

        switch(t)
        {
           case KY_BACK:
           case KY_DELETE:
           case 'Q':
             return 8;

           case 'N':
        //  case '1':
             ShowNext = ! ShowNext;
             PrepareNextWindow();
             break;

           case '3':
           case 'G':         /* grid */
             ShowGrid = !ShowGrid;
             PrepareShowGrid(figure, x, y, wid, hei);
             break;

           case 'L':
             if (Level < 9)
                { ChangeLevel(figure, x, y, wid, hei);
//                diap1 = diaps1[Level];
//                diap2 = diaps2[Level];
//                diap3 = diaps3[Level];
                  diap = diaps[Level];
                  delthink = DelayMove>>1;
                }
             break;
/*
//         case 'J':
//           setjoyparms();
//           PrepareJoyWindow();
//           break;


//         case 'K':
//           JoyEnabled = 0;
//           PrepareJoyWindow();
//           break;
//
*/

           case KY_ESC:
           case 'P':
              PauseGame();
		      AdvTicks = Ticks + DelayMove;
              break;

/*         case 'D':
              Shell();
              break;
*/
           case 'M':            /* Toggle background music */
              if (PlayListFileName)
              { MusicEnabled = !MusicEnabled;
                PrepareMusic(MODE_FORCE);
              }
              break;

/*           case 'S':
            toggle_sound(NO_SOUND);
              toggle_sound();
     //       if (SoundType & SOUND_MUSIC) music_level(Level);
              break;
*/
            case 'B':
              ToggleBottomMode();
              break;
          }  /* switch; */


          key = -1;
          curtim = Ticks;
          if (curtim < nexthink) goto DontThink;

          if (turn < BestT)
          {  rv = (BestT-turn)*diap/6;   //4
             if (rv<=1 || ranval%rv == 0)
             { rot_left(figure, figure1);
               ShowJoyPos(JOY_UP - JOY_ARROWS);
               if (test_posit (figure1, x, y, hei, wid))
               {  cop_figure(figure, figure1);
                  t = wid; wid = hei; hei = t;
                  shif = 1;
                  turn += 1;
                  key = 2;
               }
               else
                 turn = BestT;
             }
          }
          else
          if (BestX != x)
          {  rv = BestX - x;
             if (rv<0) rv = -rv;
             rv = rv*diap /8;  // 6
             if (rv <=1 || ranval % rv == 0)
             {  wid1 = (BestX > x) ? 1 : -1;
                t = x+wid1;
                if (test_posit (figure, t, y, wid, hei))
                { x = t; shif = 1;
                  key = (wid1 > 0)? JOY_RIGHT : JOY_LEFT;
                  ShowJoyPos(key - JOY_ARROWS);
                }
                else
                 BestX = x;
             }
          }
          else
          if (turn >= BestT && BestX == x)
          {
             if (!status)
                { status = 1; robthink(status);}

             if (ranval % diap < 3)    // 4  < 2
             { //if (SoundType & SOUND_BEEP)
               //           music_start(BeepDrop, 1);
               DrawBoard(Board, x, x+wid, y, y+hei);
	       y1 = drop_figure(figure, x, y, wid, hei); 
	       if (BestP == 0)	
           {  y = y1;  ok = 0; 
              key = JOY_DOWN;
	          robthink(2);
           }
	       else 
           if (y<y1-2)
	       {  downc++; shif = 1;
              key = JOY_DOWN;
           }
           else
           if (y==y1)
           {  x+= BestP; shif = 1;
              key = (BestP > 0)? JOY_RIGHT : JOY_LEFT;
			  BestX = x;  BestP = 0;
    	      robthink(2);
           }

	       ShowJoyPos(key - JOY_ARROWS);
		   adjusteye(x, y);
           key = 4;
        }
             else
               ShowJoyPos(0);
          }

          nexthink = Ticks + delthink;

    DontThink:
           curtim = Ticks;

           if (ok)
           { if (curtim >= AdvTicks)
             {  AdvTicks += DelayMove; downc++; }

	     while(--downc>=0)
             {
                ok =  down_figure(figure, x, y, wid, hei);
                if(!ok) break;
                y++; shif = 1;
             }
	     	
           } /* if (ok) */

           if (shif)
             shift_figure(figure, x, y, wid, hei);

           if (JoyEnabled)
           { if (curtim > AdvJoy)
              { if (key < 0) key = 0;
                AdvJoy += DelayJoy;
              }

              if (key >=0)  ShowJoyPos(key);
           }

          ranval += 313;
     } while(ok);



/*   After: */

     embed_figure(Board, figure, x, y, wid, hei);
     DrawBoard(Board, 0, BRD_WIDTH, y, y+hei);

     t = squeeze(curtim);
#ifdef X11 
    if (BottomMode == B_SHADOW)
                    ShowShadow(NULL, 0);
#endif
     Score += Bonus + 50*t;

     Moves++;
     if (Level < 9 && Moves >= NewLevelAt)
     {
         ChangeLevel(figure, -1, -1, -1, -1);
         diap = diaps[Level];
     }

     UpdateStatus();

     return 0;
}



